# mquery
`mquery`, short for (M)etrics Query, is a Rust library for handling PROMQL (and possibly MetricsQL) queries.

The library internally uses `prometheus-http-query` crate. As the name suggests, mquery is mainly focused on dealing with metrics queries and the resulting data, and nothing else. The aforementioned library is better suited when functionality related to the full HTTP API offered by Prometheus and other compatible solutions is required.

The library supports raw queries (strings) as well as typed queries (builder API). The builder API is provided through the type `Metric` and `Scalar` types found in the `query` module. Note that the builder API doesn't prevent you from writing invalid queries, it only helps you avoid errors to a certain degree.

Any additional utilities related to handling and processing the query data can be found in the `utils` module which is placed under the `utils` feature flag.

A compile time query validation like the one found in the `sqlx` crate is on the roadmap.

## Usage Example
```rust
use mquery::{Auth, QueryManager, query};

#[tokio::main]
async fn main() {
    dotenv::dotenv().expect("No .env file found in working dir");
    let url = std::env::var("VM_URL").expect("VM URL not found in env");
    let token = std::env::var("VM_TOKEN").expect("VM token not found in env");

    let query = query::Metric::new("total_http_requests").label_eq("method", "get");

    let _response = QueryManager::new(url.parse().unwrap())
                      .auth(Auth::Bearer(token))
                      .query(query)
                      .await
                      .expect("operation failed");
}
```
# Debugging
When you keep getting the error `Client(ClientError("failed to parse response from server due to invalid media type"))` and cannot figure out what might be wrong, it is likely that the server URL is the culprit. Ensure that only the base URL to the server is passed to the `QueryManager` which shouldn't contain the API endpoint path like `/api/v1/query`.

# Testing
Requires a local Victoria Metrics server running at `http://127.0.0.1:8428/` for the integration tests to work.

# Roadmap
- [x] Basic raw queries (Instant and Ranged)
- [x] Query Builder
  - [x] Basic queries
  - [x] Operators
    - [x] Binary Operators
    - [x] Vector matching
    - [x] Aggregation operators (implemented as functions)
  - [x] Functions (initial support for most important fns)
- [ ] Runtime syntax checking (?)
- [x] Compile time syntax checking
- [x] ~~Deserialize~~ Convert to custom data types
- [ ] ORM-like API to build queries using user defined types (derive macro)
- [ ] Processing multiple queries at once
